from pettingzoo import AECEnv
from gym import spaces
import rlcard
import random
import numpy as np
from pettingzoo.utils import wrappers


class RLCardBase(AECEnv):

    metadata = {'render.modes': ['human']}

    def __init__(self, name, num_players, obs_shape):
        super().__init__()
        self.name = name
        self.env = rlcard.make(name)
        if not hasattr(self, "agents"):
            self.agents = [f'player_{i}' for i in range(num_players)]
        self.num_agents = len(self.agents)

        dtype = self.env.reset()[0]['obs'].dtype
        if dtype == np.dtype(np.int64):
            self._dtype = np.dtype(np.int8)
        elif dtype == np.dtype(np.float64):
            self._dtype = np.dtype(np.float32)
        else:
            self._dtype = dtype

        self.observation_spaces = self._convert_to_dict([spaces.Box(low=0.0, high=1.0, shape=obs_shape, dtype=self._dtype) for _ in range(self.num_agents)])
        self.action_spaces = self._convert_to_dict([spaces.Discrete(self.env.game.get_action_num()) for _ in range(self.num_agents)])

    def seed(self, seed=None):
        self.env = rlcard.make(self.name, config={"seed": seed})

    def _scale_rewards(self, reward):
        return reward

    def _int_to_name(self, ind):
        return self.agents[ind]

    def _name_to_int(self, name):
        return self.agents.index(name)

    def _convert_to_dict(self, list_of_list):
        return dict(zip(self.agents, list_of_list))

    def observe(self, agent):
        obs = self.env.get_state(self._name_to_int(agent))
        return obs['obs'].astype(self._dtype)

    def step(self, action, observe=True):
        obs, next_player_id = self.env.step(action)
        next_player = self._int_to_name(next_player_id)
        self._last_obs = obs['obs']
        if self.env.is_over():
            self.rewards = self._convert_to_dict(self._scale_rewards(self.env.get_payoffs()))
            self.infos[next_player]['legal_moves'] = []
            self.dones = self._convert_to_dict([True if self.env.is_over() else False for _ in range(self.num_agents)])
        else:
            self.infos[next_player]['legal_moves'] = obs['legal_actions']
        self.agent_selection = next_player
        if observe:
            return obs['obs'].astype(self._dtype) if obs else self._last_obs.astype(self._dtype)

    def reset(self, observe=True):
        obs, player_id = self.env.reset()
        self.agent_selection = self._int_to_name(player_id)
        self.rewards = self._convert_to_dict([0 for _ in range(self.num_agents)])
        self.dones = self._convert_to_dict([False for _ in range(self.num_agents)])
        self.infos = self._convert_to_dict([{'legal_moves': []} for _ in range(self.num_agents)])
        self.infos[self._int_to_name(player_id)]['legal_moves'] = obs['legal_actions']
        self._last_obs = obs['obs']
        if observe:
            return obs['obs'].astype(self._dtype)
        else:
            return

    def render(self, mode='human'):
        raise NotImplementedError()

    def close(self):
        pass



from pettingzoo.utils.depreciated_module import DepreciatedModule

double_dunk_v0 = DepreciatedModule("double_dunk", "v0", "v1")
entombed_competitive_v0 = DepreciatedModule("entombed_competitive", "v0", "v1")
joust_v0 = DepreciatedModule("joust", "v0", "v1")
mario_bros_v0 = DepreciatedModule("mario_bros", "v0", "v1")
maze_craze_v0 = DepreciatedModule("maze_craze", "v0", "v1")
othello_v0 = DepreciatedModule("othello", "v0", "v1")
pong_quadrapong_v0 = DepreciatedModule("pong_quadrapong", "v0", "v1")
tennis_v0 = DepreciatedModule("tennis", "v0", "v1")
video_checkers_v0 = DepreciatedModule("video_checkers", "v0", "v1")
warlords_v0 = DepreciatedModule("warlords", "v0", "v1")
wizard_of_wor_v0 = DepreciatedModule("wizard_of_wor", "v0", "v1")



from pettingzoo.utils.to_parallel import ParallelEnv
import gym
from gym.spaces import Box,Space,Discrete
from .adv_transforms.frame_stack import stack_obs_space,stack_init,stack_obs
from .adv_transforms.frame_skip import check_transform_frameskip
from .adv_transforms.obs_delay import Delayer
import numpy as np

class ParallelWraper(ParallelEnv):
    def __init__(self, env):
        self.env = env
        self.observation_spaces = env.observation_spaces
        self.action_spaces = env.action_spaces
        self.agents = env.agents
        self.num_agents = env.num_agents

    def reset(self):
        return self.env.reset()

    def step(self, actions):
        return self.env.step(actions)

    def render(self, mode="human"):
        return self.env.render(mode)

    def close(self):
        return self.env.close()

class ObservationWrapper(ParallelWraper):
    def reset(self):
        obss = self.env.reset()
        return {agent: self._modify_observation(agent, obs) for agent, obs in obss.items()}

    def step(self, actions):
        obss, rew, done, info = self.env.step(actions)
        obss = {agent: self._modify_observation(agent, obs) for agent, obs in obss.items()}
        return obss, rew, done, info

class frame_stack(ObservationWrapper):
    def __init__(self,env,num_frames=4):
        self.stack_size = num_frames
        super().__init__(env)
        self._check_wrapper_params()
        self.observation_spaces = {agent: stack_obs_space(space, self.stack_size) for agent, space in self.observation_spaces.items()}

    def _check_wrapper_params(self):
        assert isinstance(self.stack_size, int), "stack size of frame_stack must be an int"
        for space in self.observation_spaces.values():
            if isinstance(space, Box):
                assert 1 <= len(space.shape) <= 3, "frame_stack only works for 1,2 or 3 dimentional observations"
            elif isinstance(space, Discrete):
                pass
            else:
                assert False, "Stacking is currently only allowed for Box and Discrete observation spaces. The given observation space is {}".format(obs_space)

    def reset(self):
        self.stack = {agent: stack_init(space, self.stack_size) for agent, space in self.env.observation_spaces.items()}
        return super().reset()

    def _modify_observation(self, agent, observation):
        space = self.env.observation_spaces[agent]
        self.stack[agent] = stack_obs(self.stack[agent], observation, space, self.stack_size)
        return self.stack[agent]

class delay_observations(ObservationWrapper):
    def __init__(self, env, delay):
        super().__init__(env)
        self.delay = delay

    def _modify_observation(self, agent, obs):
        return self.delayers[agent].add(obs)

    def reset(self):
        self.delayers = {agent: Delayer(space, self.delay) for agent, space in self.observation_spaces.items()}
        return super().reset()

class frame_skip(ParallelWraper):
    def __init__(self, env, frame_skip, seed=None):
        super().__init__(env)
        self.frame_skip = check_transform_frameskip(frame_skip)
        self.np_random, seed = gym.utils.seeding.np_random(seed)

    def step(self, action):
        low, high = self.frame_skip
        num_skips = int(self.np_random.randint(low, high+1))
        total_reward = {agent: 0. for agent in self.agents}

        for x in range(num_skips):
            obs, rews, done, info = super().step(action)
            for agent, rew in rews.items():
                total_reward[agent] += rew
            if all(done.values()):
                break

        return obs, total_reward, done, info



from pettingzoo.tests import api_test,seed_test,error_tests,parallel_test
from pettingzoo.mpe import simple_push_v0,simple_world_comm_v0
#from pettingzoo.sisl import multiwalker

from supersuit import frame_stack, pad_action_space, frame_skip
import numpy as np

def test_pettinzoo_frame_stack():
    _env = simple_push_v0.env()
    wrapped_env = frame_stack(_env)
    api_test.api_test(wrapped_env)

def test_pettinzoo_frame_skip():
    # this is the actual frame_skip test
    env = simple_push_v0.raw_env()
    env = frame_skip(env, 3)
    env.reset()
    for x in range(10):
        assert env.env.env.steps == (x//2)*3
        action = env.action_spaces[env.agent_selection].sample()
        next_obs = env.step(action)

def test_pettinzoo_pad_actino_space():
    _env = simple_world_comm_v0.env()
    wrapped_env = pad_action_space(_env)
    api_test.api_test(wrapped_env)
    seed_test.seed_test(lambda seed=None: pad_action_space(simple_world_comm_v0.env(seed=seed)))

def test_pettingzoo_parallel_env():
    _env = simple_world_comm_v0.parallel_env()
    wrapped_env = pad_action_space(_env)
    parallel_test.parallel_play_test(wrapped_env)



from ._mpe_utils.simple_env import SimpleEnv, make_env
from .scenarios.simple_tag import Scenario
from pettingzoo.utils.to_parallel import parallel_wrapper_fn


class raw_env(SimpleEnv):

    def __init__(self, num_good=1, num_adversaries=3, num_obstacles=2, max_frames=100):
        scenario = Scenario()
        world = scenario.make_world(num_good, num_adversaries, num_obstacles)
        super().__init__(scenario, world, max_frames)


env = make_env(raw_env)
parallel_env = parallel_wrapper_fn(env)



import random


def render_test(env):
    render_modes = env.metadata.get('render.modes')
    assert render_modes is not None, "Environment's that support rendering must define render modes in metadata"
    env.reset(observe=False)
    for mode in render_modes:
        for agent in env.agent_iter(env.num_agents*5):
            reward, done, info = env.last()
            if not done and 'legal_moves' in env.infos[agent]:
                action = random.choice(env.infos[agent]['legal_moves'])
            else:
                action = env.action_spaces[agent].sample()
            env.step(action, observe=False)
            env.render(mode=mode)
        env.reset()



from pettingzoo.utils.depreciated_module import DepreciatedModule

prison_v0 = DepreciatedModule("prison", "v0", "v1")
prospector_v0 = DepreciatedModule("prospector", "v0", "v1")
knights_archers_zombies_v0 = DepreciatedModule("knights_archers_zombies", "v0", "v1")



import timeit
import numpy as np
print(timeit.timeit('np.array(a)', "import numpy as np; a = [1,2,3,4,5,6,7,8,9,10]*10", number=10000))
print(timeit.timeit('np.array(a,dtype=np.float64)', "import numpy as np; a = [1,2,3,4,5,6,7,8,9,10]*10", number=10000))
print(timeit.timeit('np.stack(a)', "import numpy as np; a = [1,2,3,4,5,6,7,8,9,10]*10", number=10000))




def parallel_play_test(par_env):
    obs = par_env.reset()
    assert isinstance(obs, dict)
    assert set(obs.keys()).issubset(set(par_env.agents))

    for i in range(1000):
        actions = {agent:space.sample() for agent, space in par_env.action_spaces.items()}
        obs, rew, done, info = par_env.step(actions)
        assert isinstance(obs, dict)
        assert isinstance(rew, dict)
        assert isinstance(done, dict)
        assert isinstance(info, dict)
        assert set(obs.keys()).issubset(set(par_env.agents))
        assert set(rew.keys()).issubset(set(par_env.agents))
        assert set(done.keys()).issubset(set(par_env.agents))
        assert set(info.keys()).issubset(set(par_env.agents))
        if all(done.values()):
            break



from pettingzoo.utils import agent_selector
from pettingzoo import AECEnv


class _parallel_env_wrapper(AECEnv):

    metadata = {'render.modes': ['human']}

    def __init__(self, parallel_env):
        self.env = parallel_env
        self.agents = self.env.agents
        self.num_agents = len(self.agents)

        self.action_spaces = self.env.action_spaces
        self.observation_spaces = self.env.observation_spaces

    def seed(self, seed=None):
        self.env.seed(seed)

    def reset(self, observe=True):
        self._actions = {agent: None for agent in self.agents}

        self._live_agents = self.agents[:]
        self._agent_selector = agent_selector(self._live_agents)
        self.agent_selection = self._agent_selector.reset()
        self.dones = {agent: False for agent in self.agents}
        self.infos = {agent: {} for agent in self.agents}
        self.rewards = {agent: 0 for agent in self.agents}

        self._observations = self.env.reset()

        return self.observe(self.agent_selection) if observe else None

    def observe(self, agent):
        return self._observations[agent]

    def step(self, action, observe=True):
        self._actions[self.agent_selection] = action
        if self._agent_selector.is_last():
            obss, rews, dones, infos = self.env.step(self._actions)

            self._observations = obss
            self.dones = dones
            self.infos = infos
            self.rewards = rews

            self._live_agents = [agent for agent in self.agents if not dones[agent]]
            if len(self._live_agents):
                self._agent_selector = agent_selector(self._live_agents)
                self.agent_selection = self._agent_selector.reset()
        else:
            self.agent_selection = self._agent_selector.next()
        return self.observe(self.agent_selection) if observe else None

    def render(self, mode="human"):
        self.env.render()

    def close(self):
        self.env.close()



import gym
import numpy as np
import pickle

def vec_env_args(env, num_envs):
    def env_fn():
        return pickle.loads(pickle.dumps(env))
    return [env_fn]*num_envs, env.observation_space, env.action_space

def gym_vec_env(env, num_envs, multiprocessing=False):
    args = vec_env_args(env, num_envs)
    constructor = gym.vector.AsyncVectorEnv if multiprocessing else gym.vector.SyncVectorEnv
    return constructor(*args)

def stable_baselines_vec_env(env, num_envs, multiprocessing=False):
    import stable_baselines
    args = vec_env_args(env, num_envs)[:1]
    constructor = stable_baselines.common.vec_env.SubprocVecEnv if multiprocessing else stable_baselines.common.vec_env.DummyVecEnv
    return constructor(*args)

def stable_baselines3_vec_env(env, num_envs, multiprocessing=False):
    import stable_baselines3
    args = vec_env_args(env, num_envs)[:1]
    constructor = stable_baselines3.common.vec_env.SubprocVecEnv if multiprocessing else stable_baselines3.common.vec_env.DummyVecEnv
    return constructor(*args)



from gym.spaces import Discrete, Box
import numpy as np
import warnings
import magent
from pettingzoo import AECEnv
import math
from pettingzoo.magent.render import Renderer
from pettingzoo.utils import agent_selector
from .magent_env import magent_parallel_env, make_env
from pettingzoo.utils._parallel_env import _parallel_env_wrapper
from pettingzoo.utils.to_parallel import parallel_wrapper_fn
from gym.utils import EzPickle


map_size = 45
max_frames_default = 1000


def parallel_env(max_frames=max_frames_default, **reward_args):
    return _parallel_env(map_size, reward_args, max_frames)


def raw_env(max_frames=max_frames_default, **reward_args):
    return _parallel_env_wrapper(_parallel_env(map_size, reward_args, max_frames))


env = make_env(raw_env)


def load_config(map_size, step_reward=-0.01, dead_penalty=-0.1, attack_penalty=-1, attack_opponent_reward=2):
    gw = magent.gridworld
    cfg = gw.Config()

    cfg.set({"map_width": map_size, "map_height": map_size})
    cfg.set({"minimap_mode": True})

    cfg.set({"embedding_size": 10})

    options = {
        'width': 1, 'length': 1, 'hp': 10, 'speed': 1,
        'view_range': gw.CircleRange(6), 'attack_range': gw.CircleRange(1),
        'damage': 2, 'step_recover': 0.1, 'attack_in_group': True,
        'step_reward': step_reward, 'dead_penalty': dead_penalty, 'attack_penalty': attack_penalty,
    }

    melee = cfg.register_agent_type(
        "melee",
        options
    )

    options = {
        'width': 1, 'length': 1, 'hp': 3, 'speed': 2,
        'view_range': gw.CircleRange(6), 'attack_range': gw.CircleRange(2),
        'damage': 2, 'step_recover': 0.1, 'attack_in_group': True,
        'step_reward': step_reward, 'dead_penalty': dead_penalty, 'attack_penalty': attack_penalty,
    }

    ranged = cfg.register_agent_type(
        "ranged",
        options
    )

    g0 = cfg.add_group(melee)
    g1 = cfg.add_group(ranged)
    g2 = cfg.add_group(melee)
    g3 = cfg.add_group(ranged)

    arm0_0 = gw.AgentSymbol(g0, index='any')
    arm0_1 = gw.AgentSymbol(g1, index='any')
    arm1_0 = gw.AgentSymbol(g2, index='any')
    arm1_1 = gw.AgentSymbol(g3, index='any')

    # reward shaping
    cfg.add_reward_rule(gw.Event(arm0_0, 'attack', arm1_0), receiver=arm0_0, value=attack_opponent_reward)
    cfg.add_reward_rule(gw.Event(arm0_0, 'attack', arm1_1), receiver=arm0_0, value=attack_opponent_reward)
    cfg.add_reward_rule(gw.Event(arm0_1, 'attack', arm1_0), receiver=arm0_1, value=attack_opponent_reward)
    cfg.add_reward_rule(gw.Event(arm0_1, 'attack', arm1_1), receiver=arm0_1, value=attack_opponent_reward)

    cfg.add_reward_rule(gw.Event(arm1_0, 'attack', arm0_0), receiver=arm1_0, value=attack_opponent_reward)
    cfg.add_reward_rule(gw.Event(arm1_0, 'attack', arm0_1), receiver=arm1_0, value=attack_opponent_reward)
    cfg.add_reward_rule(gw.Event(arm1_1, 'attack', arm0_0), receiver=arm1_1, value=attack_opponent_reward)
    cfg.add_reward_rule(gw.Event(arm1_1, 'attack', arm0_1), receiver=arm1_1, value=attack_opponent_reward)

    # kill reward
    cfg.add_reward_rule(gw.Event(arm0_0, 'kill', arm1_0), receiver=arm0_0, value=100)
    cfg.add_reward_rule(gw.Event(arm0_0, 'kill', arm1_1), receiver=arm0_0, value=100)
    cfg.add_reward_rule(gw.Event(arm0_1, 'kill', arm1_0), receiver=arm0_1, value=100)
    cfg.add_reward_rule(gw.Event(arm0_1, 'kill', arm1_1), receiver=arm0_1, value=100)

    cfg.add_reward_rule(gw.Event(arm1_0, 'kill', arm0_0), receiver=arm1_0, value=100)
    cfg.add_reward_rule(gw.Event(arm1_0, 'kill', arm0_1), receiver=arm1_0, value=100)
    cfg.add_reward_rule(gw.Event(arm1_1, 'kill', arm0_0), receiver=arm1_1, value=100)
    cfg.add_reward_rule(gw.Event(arm1_1, 'kill', arm0_1), receiver=arm1_1, value=100)

    return cfg


def generate_map(env, map_size, handles):
    width = map_size
    height = map_size

    init_num = map_size * map_size * 0.04

    gap = 3
    # left
    n = init_num
    side = int(math.sqrt(n)) * 2
    pos = [[], []]
    ct = 0
    for x in range(width // 2 - gap - side, width // 2 - gap - side + side, 2):
        for y in range((height - side) // 2, (height - side) // 2 + side, 2):
            pos[ct % 2].append([x, y])
        ct += 1
    env.add_agents(handles[0], method="custom", pos=pos[0])
    env.add_agents(handles[1], method="custom", pos=pos[1])

    # right
    n = init_num
    side = int(math.sqrt(n)) * 2
    pos = [[], []]
    ct = 0
    for x in range(width // 2 + gap, width // 2 + gap + side, 2):
        for y in range((height - side) // 2, (height - side) // 2 + side, 2):
            pos[ct % 2].append([x, y])
        ct += 1
    env.add_agents(handles[2], method="custom", pos=pos[0])
    env.add_agents(handles[3], method="custom", pos=pos[1])


class _parallel_env(magent_parallel_env, EzPickle):
    def __init__(self, map_size, reward_args, max_frames):
        EzPickle.__init__(self, map_size, reward_args, max_frames)
        env = magent.GridWorld(load_config(map_size, **reward_args))
        names = ["redmelee", "redranged", "bluemele", "blueranged"]
        super().__init__(env, env.get_handles(), names, map_size, max_frames)

    def generate_map(self):
        generate_map(self.env, self.map_size, self.handles)
